#include <stdio.h>
#include <globus_common.h>
#include <globus_duroc_runtime.h>
#include <globus_duroc_bootstrap.h>
#include <globus_gram_myjob.h>
#include <globus_io.h>
#include "NICK_duroc_bootstrap.h"

/* inter- and intra-subjob message tags */
#define SUBJOB_MASTER_TO_SUBJOB0_MASTER_T "subjob mstr to subjob0 mstr topology"
#define SUBJOB0_MASTER_TO_SUBJOB_MASTER_T "subjob0 mstr to subjob mstr topology"
#define SUBJOB_MASTER_TO_SLAVE_T          "subjob mstr to slave topology"

#define SUBJOB_SLAVE_TO_MASTER_D          "subjob slave to master data"
#define SUBJOB_MASTER_TO_SUBJOB_MASTER_D  "subjob master to subjob master data"
#define SUBJOB_MASTER_TO_SLAVE_D          "subjob master to slave data"

#define MY_GRANK_STR_SIZE 10  /* aka NFUF :-) */


/***************************************/
/* local utility function declarations */
/***************************************/

static void get_topology(int rank_in_my_subjob, 
			int *my_subjob_size, 
			int **subjob_addresses,
			int *nprocs, 
			int *nsubjobs, 
			int *my_grank);

static void distribute_byte_array(globus_byte_t *inbuff,
				    int inbufflen,
				    int rank_in_my_subjob,
				    int my_subjob_size,
				    int *subjob_addresses,
				    int nprocs,
				    int nsubjobs,
				    int my_grank,
				    globus_byte_t **outbuffs,
				    int *outbufflens);

static void copy_byte_array_from_tagged_buff(char *tagged_intrabuff, 
					    int bufflen, 
					    char **outbuffs, 
					    int *outbufflens,
					    int *g_rank);

/*
 * NICK_globus_duroc_bootstrap_all_to_all_distribute_bytearray(
 *                                        globus_byte_t *inbuff, 
 *                                        int inbufflen, 
 *                                        int *nprocs,
 *                                        int *my_grank,
 *                                        globus_byte_t ***outbuffs,
 *                                        int **outbufflens);
 *
 * MUST be called by EVERY proc, each supplying inbuff,inbufflen
 * (may be NULL,0).
 *
 * fills the last 4 args as follows:
 *     nprocs  - total number of procs 
 *     my_grank - unique id in range [0, nprocs-1]. 
 *                preserves order of your subjob as it appears in envoking RSL,
 *                i.e., my_rank will be greater than all procs of 
 *                subjobs appearing before mine in envoking RSL and 
 *                less than all procs of subjobs appearing after mine.
 *     outbuffs - malloc'd a vector (length nprocs) where each element
 *                points to a (also malloc'd) copy of the byte array 
 *                supplied by a particular proc such that 
 *                outbuffs[i] is a copy of the byte array of
 *                the proc assigned my_grank=i (each proc should
 *                find a copy of the byte array they supplied in
 *                outbuffs[my_grank])
 *     outbufflens - malloc'd a vector (length nprocs) of ints where
 *                   outbufflens[i] is the length of the byte array
 *                   outbuffs[i].
 *                
 */

/* NICK: should return 'globus-like' rc ... not 'void' */

void NICK_globus_duroc_bootstrap_all_to_all_distribute_bytearray(
					    globus_byte_t *inbuff, 
					    int inbufflen, 
					    int *nprocs,
					    int *my_grank,
					    globus_byte_t ***outbuffs,
					    int **outbufflens)
{
    int rank_in_my_subjob;
    int my_subjob_size;    /* used by subjobmasters only */
    int nsubjobs;          /* used by subjobmasters only */
    int *subjob_addresses; /* used by subjobmasters only */

    globus_module_activate(GLOBUS_DUROC_BOOTSTRAP_MODULE);

    globus_duroc_runtime_intra_subjob_rank(&rank_in_my_subjob);

    get_topology(rank_in_my_subjob, 
		&my_subjob_size,
		&subjob_addresses,
		nprocs, 
		&nsubjobs,
		my_grank);

    /* NICK: exiting on these failed mallocs is not right thing */
    /*       to do.  should set error rc and return. fix that   */
    /*       later when i learn more about globus rc stuff.     */
    if (!(*outbuffs = 
	(globus_byte_t **) globus_libc_malloc(*nprocs*sizeof(globus_byte_t *))))
    {
	globus_libc_fprintf(stderr, 
	    "ERROR: failed malloc of %ld bytes\n", 
	    *nprocs*sizeof(globus_byte_t *));
	exit(1);
    } /* endif */
    if (!(*outbufflens = (int *) globus_libc_malloc(*nprocs*sizeof(int))))
    {
	globus_libc_fprintf(stderr, 
	    "ERROR: failed malloc of %ld bytes\n", *nprocs*sizeof(int));
	exit(1);
    } /* endif */

    distribute_byte_array(inbuff,
			inbufflen,
			rank_in_my_subjob,
			my_subjob_size,
			subjob_addresses,
			*nprocs,
			nsubjobs,
			*my_grank,
			*outbuffs,
			*outbufflens);

    /************/
    /* clean-up */
    /************/

    if (!rank_in_my_subjob)
	/* subjob master */
	globus_libc_free(subjob_addresses);

    globus_module_deactivate(GLOBUS_DUROC_BOOTSTRAP_MODULE);

} /* end NICK_globus_duroc_bootstrap_all_to_all_distribute_bytearray() */

/***************************/
/* Local Utility Functions */
/***************************/

/*
 * get_topology(int rank_in_my_subjob,
 *              int *my_subjob_size,
 *              int **subjob_addresses,
 *              int *nprocs,
 *              int *nsubjobs,
 *              int *my_grank)
 *
 * MUST be called by EVERY proc, each supplying rank_in_my_subjob
 * rank_in_my_subjob==0 -> subjobmaster, else subjobslave
 *
 * fills the remaining args:
 *     my_subjob_size  - populated for subjobmasters only, 
 *                       nprocs (including master) in their subjob 
 *     subjob_addresses - malloc'd and filled for OTHER subjobmasters only
 *                        inter_subjob_addr's of other subjobmasters
 *                        my subjob_addr NOT included (so njobs-1)
 *     nsubjobs - populated for subjobmasters only.
 *                total number of procs
 *     nprocs - total number of procs
 *     my_grank - my rank in [0, nprocs-1]
 *     
 */

/* NICK: should return 'globus-like' rc ... not 'void' */

static void get_topology(int rank_in_my_subjob, 
			int *my_subjob_size, 
			int **subjob_addresses,
			int *nprocs, 
			int *nsubjobs, 
			int *my_grank)
{
    char topology_buff[GRAM_MYJOB_MAX_BUFFER_LENGTH];
    char *buff;
    int bufflen;
    int i;
    int sj0_master_idx;    /* used by subjobmasters only */
    int *job_sizes;        /* used by subjobmaster 0 only */
    int *g_ranks;          /* used by subjobmaster 0 only */

    if (rank_in_my_subjob)
    {
	/* subjob slave */

	globus_duroc_runtime_intra_subjob_receive(
	    SUBJOB_MASTER_TO_SLAVE_T,         /* tag */
	    &bufflen,                         /* nbytes received? */
	    (globus_byte_t *) topology_buff);  /* message */

	sscanf(topology_buff, "%d %d", nprocs, my_grank);
    }
    else
    {
	/* subjob master */
	int duroc_subjobmaster_rank;
	int my_subjob_addr;
	int rsl_subjob_rank;
	char *rsl_subjob_rank_env_var;

	globus_duroc_runtime_intra_subjob_size(my_subjob_size);
	globus_duroc_runtime_inter_subjob_structure(&my_subjob_addr,
						    nsubjobs,
						    subjob_addresses);
	/* finding index of master subjob 0, he */
	/* is the one with the lowest address   */
	for (i = 0, sj0_master_idx = -1, duroc_subjobmaster_rank = 0; 
	    i < *nsubjobs; i ++)
	{
	    if ((sj0_master_idx == -1 && (*subjob_addresses)[i] < my_subjob_addr)
	    || (sj0_master_idx != -1 
		    && (*subjob_addresses)[i] < (*subjob_addresses)[sj0_master_idx]))
		sj0_master_idx = i;
	    if ((*subjob_addresses)[i] < my_subjob_addr)
		duroc_subjobmaster_rank ++;
	} /* endfor */
	/* globus_duroc_runtime_inter_subjob_structure reports the */
	/* number of REMOTE subjobs (*other* than the subjob i'm   */
	/* master of).  to get the TOTAL number of subjobs in this */
	/* run i must increment the value reported by              */
	/* globus_duroc_runtime_inter_subjob_structure             */
	(*nsubjobs) ++;

	/* NICK: should not exit here ... should set globus-like rc. */
	if (!(rsl_subjob_rank_env_var=getenv("GLOBUS_DUROC_SUBJOB_INDEX")))
	{
	    globus_libc_fprintf(stderr, 
"ERROR: required environment variable GLOBUS_DUROC_SUBJOB_INDEX not set.\n");
	    globus_libc_fprintf(stderr, 
"       Each subjob in envoking RSL must have GLOBUS_DUROC_SUBJOB_INDEX\n");
	    globus_libc_fprintf(stderr, 
"       set to rank (0, 1, 2, ...) of subjob as it appears in the envoking RSL.\n");
	    exit(1);
	} /* endif */
	rsl_subjob_rank = atoi(rsl_subjob_rank_env_var);
	if (rsl_subjob_rank < 0 || rsl_subjob_rank >= *nsubjobs)
	{
	    globus_libc_fprintf(stderr, 
	"ERROR: env variable GLOBUS_DUROC_SUBJOB_INDEX %d must be >= 0 and\n", 
		rsl_subjob_rank);
	    globus_libc_fprintf(stderr, 
		"ERROR: less than the number of subjobs %d for this run.\n", 
		*nsubjobs);
	    exit(1);
	} /* endif */

	if (duroc_subjobmaster_rank)
	{
	    /* NOT master of subjob 0 */

	    sprintf(topology_buff,
		    "%d %d %d",
		    duroc_subjobmaster_rank,
		    rsl_subjob_rank,
		    *my_subjob_size);

	    globus_duroc_runtime_inter_subjob_send(
		(*subjob_addresses)[sj0_master_idx],  /* dest */
		SUBJOB_MASTER_TO_SUBJOB0_MASTER_T, /* tag */
		strlen(topology_buff)+1,            /* nbytes */
		(globus_byte_t *) topology_buff);  /* data */

	    globus_duroc_runtime_inter_subjob_receive(
		SUBJOB0_MASTER_TO_SUBJOB_MASTER_T, /* tag */
		&bufflen,                          /* nbytes received? */
		(globus_byte_t **) &buff);         /* message */

	    sscanf(buff, "%d %d", nprocs, my_grank);

	    globus_libc_free(buff);
	}
	else
	{
	    /* master of subjob 0 */

	    int j;
	    int temp;
	    /* vectors len nsubjobs, all indexed by duroc_subjobmaster_rank */
	    int *rsl_ranks; /* received from other subjob masters */

	    /* NICK: exiting on these failed mallocs is not right thing */
	    /*       to do.  should set error rc and return. fix that   */
	    /*       later when i learn more about globus rc stuff.     */
	    if (!(rsl_ranks = 
		(int *) globus_libc_malloc(*nsubjobs*sizeof(int))))
	    {
		globus_libc_fprintf(stderr, 
		    "ERROR: failed malloc of %ld bytes\n", 
		    *nsubjobs*sizeof(int));
		exit(1);
	    } /* endif */
	    if (!(job_sizes = 
		(int *) globus_libc_malloc(*nsubjobs*sizeof(int))))
	    {
		globus_libc_fprintf(stderr, 
		    "ERROR: failed malloc of %ld bytes\n", 
		    *nsubjobs*sizeof(int));
		exit(1);
	    } /* endif */
	    if (!(g_ranks = (int *) globus_libc_malloc(*nsubjobs*sizeof(int))))
	    {
		globus_libc_fprintf(stderr, 
		    "ERROR: failed malloc of %ld bytes\n", 
		    *nsubjobs*sizeof(int));
		exit(1);
	    } /* endif */

	    /* need to sort subjob_addresses so that i may associate */
	    /* (using incoming duroc_subjobmaster_rank) incoming     */
	    /* rsl_subjob_rank and my_subjob_size with dest addr     */
	    /* in subjob_addresses                                   */
	    for (i = 1; i < *nsubjobs-1; i ++)
	    {
		for (j = i; j > 0; j --)
		{
		    if ((*subjob_addresses)[j] < (*subjob_addresses)[j-1])
		    {
			temp = (*subjob_addresses)[j];
			(*subjob_addresses)[j] = (*subjob_addresses)[j-1];
			(*subjob_addresses)[j-1] = temp;
		    } /* endif */
		} /* endfor */
	    } /* endfor */

	    /* rsl_ranks[] and job_sizes[] are indexed by  */
	    /* duroc_subjobmaster_rank, and i know that my */
	    /* duroc_subjobmaster_rank==0                  */
	    rsl_ranks[0] = rsl_subjob_rank;
	    job_sizes[0] = *my_subjob_size;

	    for (i = 1; i < *nsubjobs; i ++)
	    {
		int ranks, sizes;
		
		/*
		 * receiving 3 longs from other subjob master
                 *    duroc_subjobmaster_rank (used to index job_sizes[] 
                 *                             and rsl_ranks[])
                 *    rsl_subjob_rank 
                 *    my_subjob_size
                 */
		globus_duroc_runtime_inter_subjob_receive(
		    SUBJOB_MASTER_TO_SUBJOB0_MASTER_T, /* tag */
		    &bufflen,                          /* nbytes received? */
		    (globus_byte_t **) &buff);         /* message */

		sscanf(buff, "%d %d %d", &j, &ranks, &sizes);
		rsl_ranks[j] = ranks;
		job_sizes[j] = sizes;
		
		globus_libc_free(buff);
	    } /* endfor */

	    /* calculating nprocs and everyones' g_rank based */
	    /* on rsl_rank and job_sizes ...                  */
	    /* mygrank = sum job_size for all rsl_ranks       */
	    /*           that are less than mine              */
	    for (i = 0, *nprocs = 0; i < *nsubjobs; i ++)
	    {
		(*nprocs) += job_sizes[i];
		for (g_ranks[i] = 0, j = 0; j < *nsubjobs; j ++)
		if (rsl_ranks[i] > rsl_ranks[j])
		    g_ranks[i] += job_sizes[j];
	    } /* endfor */
	    *my_grank = g_ranks[0];

	    /* sending other subjob masters nprocs and their g_rank */
	    for (i = 0; i < *nsubjobs-1; i ++)
	    {
		sprintf(topology_buff, "%d %d", *nprocs, g_ranks[i+1]);
		globus_duroc_runtime_inter_subjob_send(
		    (*subjob_addresses)[i],               /* dest */
		    SUBJOB0_MASTER_TO_SUBJOB_MASTER_T, /* tag */
		    strlen(topology_buff)+1,           /* nbytes */
		    (globus_byte_t *) topology_buff);  /* data */
	    } /* endfor */

	    globus_libc_free(rsl_ranks);
	    globus_libc_free(job_sizes);
	    globus_libc_free(g_ranks);

	} /* endif */

	/* all subjob masters sending nprocs and g_rank to their slaves */
	for (i = 1; i < *my_subjob_size; i ++)
	{
	    sprintf(topology_buff, "%d %d", *nprocs, (*my_grank) + i);
	    globus_duroc_runtime_intra_subjob_send(
		i,                                /* dest */
		SUBJOB_MASTER_TO_SLAVE_T,         /* tag */
		strlen(topology_buff)+1,          /* nbytes */
		(globus_byte_t *) topology_buff); /* data */
	} /* endfor */
    } /* endif */

} /* end get_topology() */

static void distribute_byte_array(globus_byte_t *inbuff,
				    int inbufflen,
				    int rank_in_my_subjob,
				    int my_subjob_size,
				    int *subjob_addresses,
				    int nprocs,
				    int nsubjobs,
				    int my_grank,
				    globus_byte_t **outbuffs,
				    int *outbufflens)
{

    int i;
    int j;
    char *buff;
    int bufflen;
    char tagged_intrabuff[GRAM_MYJOB_MAX_BUFFER_LENGTH];
    int g_rank;

    /* everyone needs to copy their byte array into outbuffs */
    if (!(outbuffs[my_grank] = (globus_byte_t *) globus_libc_malloc(inbufflen)))
    {
	globus_libc_fprintf(stderr, 
	    "ERROR: failed malloc of %d bytes\n", inbufflen);
	exit(1);
    } /* endif */
    memcpy(outbuffs[my_grank], inbuff, inbufflen);
    outbufflens[my_grank] = inbufflen;

    /* tagging and copying my byte array for distribution */
    sprintf(tagged_intrabuff, "%d", my_grank);
    memcpy(tagged_intrabuff+MY_GRANK_STR_SIZE, inbuff, inbufflen);

    if (rank_in_my_subjob)
    {
	/* subjob slave */

	/* send my byte array to my master */
	globus_duroc_runtime_intra_subjob_send(
	    0,                                /* dest */
	    SUBJOB_SLAVE_TO_MASTER_D,         /* tag */
	    MY_GRANK_STR_SIZE+inbufflen,      /* nbytes */
	    (globus_byte_t *) tagged_intrabuff); /* data */
	
	/* receiving all other byte arrays from my master */
	for (i = 0; i < nprocs-1; i ++)
	{
	    globus_duroc_runtime_intra_subjob_receive(
		SUBJOB_MASTER_TO_SLAVE_D, /* tag */
		&bufflen,                 /* nbytes received? */
		(globus_byte_t *) tagged_intrabuff);        /* message */

	    copy_byte_array_from_tagged_buff(tagged_intrabuff, 
					    bufflen, 
					    (char **) outbuffs, 
					    outbufflens,
					    &g_rank);
	} /* endfor */
    }
    else 
    {
	/* subjob master */

	/* sending my byte array to all my slaves */
	for (i = 1; i < my_subjob_size; i ++)
	{
	    globus_duroc_runtime_intra_subjob_send(
		i,                                /* dest */
		SUBJOB_MASTER_TO_SLAVE_D,         /* tag */
		MY_GRANK_STR_SIZE+inbufflen,      /* nbytes */
		(globus_byte_t *) tagged_intrabuff);  /* data */
	} /* endfor */

	/* sending my byte array to other subjob masters */
	for (i = 0; i < nsubjobs-1; i ++)
	{
	    globus_duroc_runtime_inter_subjob_send(
		subjob_addresses[i],                 /* dest */
		SUBJOB_MASTER_TO_SUBJOB_MASTER_D,    /* tag */
		MY_GRANK_STR_SIZE+inbufflen,         /* nbytes */
		(globus_byte_t *) tagged_intrabuff); /* data */
	} /* endfor */

	/* receiving byte arrays from my slaves.  making a copy */
	/* for myself, distributing it to my other slaves, and  */
	/* finally other subjob masters                         */
	for (i = 0; i < my_subjob_size-1; i ++)
	{
	    globus_duroc_runtime_intra_subjob_receive(
		SUBJOB_SLAVE_TO_MASTER_D, /* tag */
		&bufflen,                 /* nbytes received? */
		(globus_byte_t *) tagged_intrabuff);        /* message */

	    /* making copy for myself */
	    copy_byte_array_from_tagged_buff(tagged_intrabuff, 
					    bufflen, 
					    (char **) outbuffs, 
					    outbufflens,
					    &g_rank);

	    /* send to other of slaves */
	    for (j = 1; j < my_subjob_size; j ++)
	    {
		if (my_grank+j != g_rank)
		{
		    /* message originated at slave other than this one */
		    globus_duroc_runtime_intra_subjob_send(
			j,                        /* dest */
			SUBJOB_MASTER_TO_SLAVE_D, /* tag */
			bufflen,                  /* nbytes */
			(globus_byte_t *) tagged_intrabuff);        /* data */
		} /* endif */
	    } /* endfor */

	    /* sending to other subjob masters */
	    for (j = 0; j < nsubjobs-1; j ++)
	    {
		globus_duroc_runtime_inter_subjob_send(
		    subjob_addresses[j],                 /* dest */
		    SUBJOB_MASTER_TO_SUBJOB_MASTER_D,    /* tag */
		    bufflen,                             /* nbytes */
		    (globus_byte_t *) tagged_intrabuff); /* data */
	    } /* endfor */
	} /* endfor */

	/* receiving from other subjob masters, making a copy for */
	/* myself, distributing to my slaves */
	for (i = 0; i < nprocs-my_subjob_size; i ++)
	{
	    globus_duroc_runtime_inter_subjob_receive(
		SUBJOB_MASTER_TO_SUBJOB_MASTER_D, /* tag */
		&bufflen,                         /* nbytes received? */
		(globus_byte_t **) &buff);         /* message */

	    /* make copy for myself */
	    copy_byte_array_from_tagged_buff(buff, 
					    bufflen, 
					    (char **) outbuffs, 
					    outbufflens,
					    &g_rank);

	    /* distribute to all my slaves */
	    for (j = 1; j < my_subjob_size; j ++)
	    {
		globus_duroc_runtime_intra_subjob_send(
		    j,                        /* dest */
		    SUBJOB_MASTER_TO_SLAVE_D, /* tag */
		    bufflen,                  /* nbytes */
		    (globus_byte_t *) buff);                    /* data */
	    } /* endfor */

	    globus_libc_free(buff);

	} /* endfor */
    } /* endif */

} /* end distribute_byte_array() */

static void copy_byte_array_from_tagged_buff(char *tagged_intrabuff, 
					    int bufflen, 
					    char **outbuffs, 
					    int *outbufflens,
					    int *g_rank)
{
    int len;
    
    sscanf(tagged_intrabuff, "%d", g_rank);

    if (len = bufflen - MY_GRANK_STR_SIZE)
    {
	if (!(outbuffs[*g_rank] = (char *) globus_libc_malloc(len)))
	{
	    globus_libc_fprintf(stderr,
		"ERROR: failed malloc of %d bytes\n", len);
	    exit(1);
	} /* endif */

	memcpy(outbuffs[*g_rank], 
	    tagged_intrabuff + MY_GRANK_STR_SIZE,
	    len);
    }
    else
	outbuffs[*g_rank] = (char *) NULL;

    outbufflens[*g_rank] = len;

} /* end copy_byte_array_from_tagged_buff() */
